/******************************************************************************
* raster2vector.c - $Revision: 1.25 $ - $Date: 2004/11/14 19:36:30 $
*
*   Copyright (C) 2004 Craig B. Agricola (craig@theagricolas.org)
*
* This project started because I wanted a scalable PostScript version of my
* signature, which I had scanned. The tool takes an image in PNM format and
* creates PostScript code to trace over the non-white bits. This only really
* makes sense if the image input is an edge map. While writing this, it
* occurred to me that this is also a pretty good way to efficiently store
* scanned documents, assuming that you want a high resolution monochrome
* representation of the scanned page for storage. The project still has plenty
* of room for improvement. Of course, after I had finished it to its current
* state, I found potrace, which is a very very similar, though much more
* mature, project. As far as I can tell, the algorithms used are completely
* different however, so depending on the input and the desired use of the
* output, one might perform better than the other. In my limited testing, my
* raster2vector produced significantly smaller output files, which is obviously
* better for archiving scanned documents, but it was by no means an exhaustive
* or objective study.
*
* BUILDING
*
* This project requires getopt_auto.c and getopt_auto.h from my getopt_auto
* project.  They should have been distributed along with this file.  Use
* the included Makefile to build.
*
* USAGE
*
* raster2vector is a filter, so it takes the input on the standard input and
* writes the output to standard output.  All of the algorithm controls are
* controlled by standard long options, which are documented pretty well in
* the output of 'raster2vector --help'.
*
* LICENSE
*
* raster2vector is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation; either version 2, or (at your option) any later version.
*
* raster2vector is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
* details.
*
* You should be able to download a copy of the GNU General Public License from
* http://www.gnu.org/licenses/gpl.txt
* If not, write to the Free Software Foundation,
* 59 Temple Place, Suite 330, Boston, MA 02111 USA.
*
******************************************************************************/

#include <stdio.h>
#include "getopt_auto.h"

typedef unsigned char uint8;

int xoff[8] = { 1, 1, 0,-1,-1,-1, 0, 1 };
int yoff[8] = { 0,-1,-1,-1, 0, 1, 1, 1 };

struct path_point_s {
  int x;
  int y;
  struct path_point_s *next;
  struct path_point_s *prev;
};

struct path_s {
  int length;
  int closed;
  struct path_point_s *start;
  struct path_point_s *end;
};

struct path_list_s {
  struct path_s *path;
  struct path_list_s *next;
};

struct path_s *new_path(int x, int y) {
  struct path_s *p =
    (struct path_s*) malloc(sizeof(struct path_s));
  struct path_point_s *pp =
    (struct path_point_s*) malloc(sizeof(struct path_point_s));
  pp->x = x;
  pp->y = y;
  p->length = 1;
  p->closed = 0;
  p->start = p->end = pp;
  pp->next = pp->prev = 0;
  return(p);
}

void walk_path(int free_it, int visit_val,
               int width, int height, uint8 *visited,
               struct path_s *path)
{
  if (path) {
    struct path_point_s *cur, *next;
    next = path->start;
    while(cur = next) {
      if (visited) visited[cur->x + (cur->y * width)] = visit_val;
      next = cur->next;
      if (free_it) free(cur);
    }
    if (free_it) free(path);
  }
}

#define free_path(p)                    walk_path(1, 0, 0, 0, 0, p)
#define unvisit_path(w,h,v,p)           walk_path(0, 0, w, h, v, p)
#define unvisit_and_free_path(w,h,v,p)  walk_path(1, 0, w, h, v, p)
#define visit_path(w,h,v,p)             walk_path(0, 1, w, h, v, p)

void append_path_and_free_source(struct path_s *dst, struct path_s *src) {
  struct path_point_s *pp;
  if (src) {
    if (dst) {
      dst->end->next   = src->start;
      src->start->prev = dst->end;
      dst->end         = src->end;
      dst->length      = dst->length + src->length;
      src->start       = 0;
      src->end         = 0;
    }
    free_path(src);
  }
}

void free_path_list(struct path_list_s *path_list) {
  struct path_list_s *next;
  for(; path_list; path_list = next) {
    next = path_list->next;
    free_path(path_list->path);
    free(path_list);
  }
}

typedef enum { FAR, SHORT } goal_t;

struct path_s *find_path(int startx, int starty,
                         goal_t goal, int goal_endx, int goal_endy,
                         int width, int height, int depth,
                         uint8 *edge, uint8 *visited)
{
  struct path_s *ret_path = 0;
  if ((edge[startx+(starty*width)] < depth) &&
      (visited[startx+(starty*width)] == 0))
  {
    ret_path = new_path(startx, starty);
    visited[startx + (starty*width)] = 1;

    struct path_s *best_path = 0;
    int d;
    for(d=0; d<8; d++) {
      struct path_s *tpath;
      tpath = find_path(startx + xoff[d], starty + yoff[d],
                        goal, goal_endx, goal_endy,
                        width, height, depth, edge, visited);
      if (tpath) {
        if ((best_path == 0)                                           ||
            ((goal == FAR)   && (tpath->length > best_path->length))   ||
            ((goal == SHORT) && (tpath->length < best_path->length) &&
             (tpath->end->x == goal_endx) && (tpath->end->y == goal_endy))   )
        {
          unvisit_and_free_path(width, height, visited, best_path);
          best_path = tpath;
        } else
          unvisit_and_free_path(width, height, visited, tpath);
      }
    }
    append_path_and_free_source(ret_path, best_path);
    if ((goal == SHORT) && ((ret_path->end->x != goal_endx) ||
                            (ret_path->end->y != goal_endy)))
    {
      free_path(ret_path);
      ret_path = 0;
    }
  }
  return(ret_path);
}

struct path_list_s* edge_follower(int width, int height, int depth,
                                  uint8 *edge,
                                  int min_path_len)
{
  struct path_list_s *path_list = 0;
  unsigned char *visited;
  int i, x, y;

  visited = (unsigned char*) malloc ( width * height * sizeof(unsigned char) );
  for(i=width*height; i>0; i--) visited[i-1]=0;

  for(y=0; y<height; y++) {
    for(x=0; x<width; x++) {
      struct path_s *tpath;
      tpath = find_path(x, y, FAR, 0, 0, width, height, depth, edge, visited);
      if (tpath) {
        int startx = tpath->end->x;
        int starty = tpath->end->y;
        unvisit_and_free_path(width, height, visited, tpath);

        tpath = find_path(startx, starty, FAR, 0, 0,
                          width, height, depth, edge, visited);
        if (tpath) {
          int endx = tpath->end->x;
          int endy = tpath->end->y;
          unvisit_and_free_path(width, height, visited, tpath);

          tpath = find_path(startx, starty, SHORT, endx, endy, width, height,
                            depth, edge, visited);
        }
      }
      if (tpath && tpath->length > min_path_len) {
        struct path_list_s *new;
        new = (struct path_list_s*) malloc (sizeof(struct path_list_s));
        new->path = tpath;
        new->next = path_list;
        path_list = new;
      }
    }
  }
  free(visited);
  return(path_list);
}

int distance_between_points_squared(struct path_point_s *a,
                                    struct path_point_s *b)
{
  int dist_squared = 0;
  if (a && b)
    dist_squared = (((a->x) - (b->x)) * ((a->x) - (b->x))) +
                   (((a->y) - (b->y)) * ((a->y) - (b->y)));
  return(dist_squared);
}

double distance_between_point_and_line_squared(struct path_point_s *a,
                                               struct path_point_s *b,
                                               struct path_point_s *p)
{
  double dist_squared;
  int seg_len_squared = distance_between_points_squared(a, b);
  if (seg_len_squared) {
    int tds;
    tds =   (((a->y) - (b->y)) * (p->x))
          - (((a->x) - (b->x)) * (p->y))
          + (((a->x) * (b->y)) - ((a->y) * (b->x)));
    tds = tds * tds;
    dist_squared = ((double)tds) / seg_len_squared;
  } else {
    dist_squared = distance_between_points_squared(a, p);
  }
  return(dist_squared);
}

void smooth_paths(struct path_list_s *path_list,
                  double max_seg_length, double max_deviation)
{
  struct path_s *path;
  struct path_point_s *curpt, *walkpt, *walkprev, *candidate;
  struct path_list_s *path_walk;
  int state;
  double dist;

  max_seg_length = max_seg_length * max_seg_length;
  max_deviation  = max_deviation  * max_deviation;

  for(path_walk = path_list; path_walk; path_walk = path_walk->next) {
    if (path = path_walk->path) {
      for(curpt = path->start; curpt && curpt->next; curpt = curpt->next) {
        state = 0;
        candidate = 0;
        for(walkpt = path->end;
            walkpt && (walkpt != curpt);
            walkpt = walkpt->prev)
        {
          if (state == 1) {
            dist = distance_between_point_and_line_squared(curpt, candidate,
                                                           walkpt);
            if (dist > max_deviation) { state = 0; candidate = 0; }
          }
          /* Yes, the fall through to do state 0 right away is intentional... */
          if (state == 0) {
            dist = distance_between_points_squared(curpt, walkpt);
            if (dist <= max_seg_length) { state = 1; candidate = walkpt; }
          }
        }
        if (candidate && (candidate->prev != curpt)) {
          for(walkpt = candidate->prev;
              walkpt && (walkpt != curpt);
              walkpt = walkprev)
          {
            walkprev = walkpt->prev;
            free(walkpt);
          }
          curpt->next = candidate;
          candidate->prev = curpt;
        }
      }
    }
  }
}

void remove_tails(struct path_list_s *path_list,
                  int max_tail_len, double neck_width)
{
  struct path_point_s *startbound,    *endbound;
  struct path_point_s *start,         *end;
  struct path_point_s *startwalk,     *endwalk;
  struct path_point_s *bestpassstart, *bestpassend;
  struct path_point_s *beststart,     *bestend;
  int                  bestpassdist;
  int                  bestdist;
  int c;
  neck_width = neck_width * neck_width;
  for(; path_list; path_list=path_list->next) {
    beststart = bestend = 0;
    startbound = start = path_list->path->start;
    endbound   = end   = path_list->path->end;
    for(c=max_tail_len+1; c && startbound->next && endbound->prev; c--) {
      startbound = startbound->next;
      endbound   = endbound->prev;
      if ((startbound == endbound) || (startbound->next == endbound)) break;
    }
    for(startwalk=start; startwalk!=startbound; startwalk=startwalk->next) {
      bestpassstart = bestpassend = 0;
      for(endwalk=end; endwalk!=endbound; endwalk=endwalk->prev) {
        int dist = distance_between_points_squared(startwalk, endwalk);
        if ((!bestpassstart) || (dist < bestpassdist)) {
          bestpassdist = dist;
          bestpassstart = startwalk;
          bestpassend = endwalk;
        } else if (bestpassstart) endwalk = endbound->next;
      }
      if (!beststart || (bestpassdist < bestdist)) {
        bestdist = bestpassdist;
        beststart = bestpassstart;
        bestend = bestpassend;
      }
    }
    if (beststart && (bestdist <= neck_width)) {
      path_list->path->start = beststart;
      path_list->path->end   = bestend;
      for(startwalk = beststart->prev; startwalk; startwalk = startwalk->prev)
        free(startwalk);
      for(endwalk = bestend->next; endwalk; endwalk = endwalk->next)
        free(endwalk);
      beststart->prev = 0;
      bestend->next = 0;
    }
  }
}

void join_paths(struct path_list_s *path_list, double join_radius) {
  struct path_list_s *work, *walk, *walk_prev;
  join_radius = join_radius * join_radius;
  int joined;
  do {
    joined = 0;
    for(work=path_list; work && work->path; work=work->next) {
      struct path_point_s *workstart = work->path->start;
      struct path_point_s *workend   = work->path->end;
      if (distance_between_points_squared(workstart, workend) > join_radius) {
        walk_prev = work;
        for(walk=work->next; walk && walk->path; walk=walk->next) {
          struct path_point_s *walkstart = walk->path->start;
          struct path_point_s *walkend   = walk->path->end;
          if (distance_between_points_squared(walkstart, walkend) > join_radius)
          {
            int joining = 0;
            if        (distance_between_points_squared(workstart,
                                                       walkstart)<=join_radius)
            {
              struct path_point_s *copypt, *prevpt, *nextpt;
              prevpt = workstart;
              for(copypt=walkstart; copypt; copypt=nextpt) {
                nextpt = copypt->next;
                prevpt->prev = copypt;
                copypt->next = prevpt;
                prevpt = copypt;
              }
              prevpt->prev = 0;
              work->path->start = prevpt;
              joining = 1;
            } else if (distance_between_points_squared(workstart,
                                                       walkend)  <=join_radius)
            {
              walkend->next = workstart;
              workstart->prev = walkend;
              work->path->start = walkstart;
              joining = 1;
            } else if (distance_between_points_squared(workend,
                                                       walkstart)<=join_radius)
            {
              workend->next = walkstart;
              walkstart->prev = workend;
              work->path->end = walkend;
              joining = 1;
            } else if (distance_between_points_squared(workend,
                                                       walkend)  <=join_radius)
            {
              struct path_point_s *copypt, *prevpt, *nextpt;
              prevpt = workend;
              for(copypt=walkend; copypt; copypt=nextpt) {
                nextpt = copypt->prev;
                prevpt->next = copypt;
                copypt->prev = prevpt;
                prevpt = copypt;
              }
              prevpt->next = 0;
              work->path->end = prevpt;
              joining = 1;
            }
            if (joining) {
              walk_prev->next = walk->next;
              free(walk->path);
              free(walk);
              walk = walk_prev;
              joined = 1;
            }
          }
          walk_prev = walk;
        }
      }
    }
  } while (joined);
}

void close_paths(struct path_list_s *path_list, double close_radius) {
  struct path_point_s *start, *end, *newpt;
  close_radius = close_radius * close_radius;
  for(; path_list; path_list=path_list->next) {
    start = path_list->path->start;
    end   = path_list->path->end;
    int dist = distance_between_points_squared(start, end);
    if (dist <= close_radius)
      path_list->path->closed = 1;
  }
}

void sort_paths(struct path_list_s *path_list) {

}

char errstr[128];

void chop_newline(char *str) {
  char *p;
  for(p=str; (*p!=0) && (*p != '\n'); p++);
  *p = 0;
}

int write_path_list_to_ps_file(char *filename, int width, int height,
                               int for_debug, int add_comments,
                               int minimize_output_size,
                               int relative_moves, int relative_lines,
                               struct path_list_s *path_list)
{
  FILE *out;
  int path_num = 0;
  int x, y;
  int lastx, lasty;
  lastx = lasty = 0;
  char *moveto, *lineto, *closepath;
  if (filename) {
    if ((out = fopen(filename, "w"))==NULL) {
      snprintf(errstr, 128, "Failed to open PS file %s for writing.", filename);
      return(-1);
    }
  } else out = stdout;

  fprintf(out, "%!PS-Adobe-3.0 EPSF-3.0\n");
  fprintf(out, "%%%%BoundingBox: 0 0 %d %d\n", width, height);
  if (minimize_output_size) {
    fprintf(out, "/bd{bind def}bind def ");
    fprintf(out, "/m{%smoveto}bd ", relative_moves?"r":"");
    fprintf(out, "/l{%slineto}bd ", relative_lines?"r":"");
    fprintf(out, "/c{closepath}bd\n");
    moveto = "m";
    lineto = "l";
    closepath = "c";
  } else {
    moveto = relative_moves?"rmoveto":"moveto";
    lineto = relative_lines?"rlineto":"lineto";
    closepath = "closepath";
  }
  if (for_debug) {
    fprintf(out, "1 0 0 setrgbcolor\n1 setlinewidth\n");
    fprintf(out, "(Times-Roman) findfont 6 scalefont setfont\n");
  } else if (relative_moves)
    fprintf(out, "0 0 moveto\n");
  fprintf(out, "\n");

  struct path_list_s *curpl, *nextpl;
  nextpl = path_list;
  while(curpl = nextpl) {
    struct path_s *path = curpl->path;
    if (path && path->start) {
      struct path_point_s *cur = path->start;
      struct path_point_s *next;
      if (add_comments) {
        fprintf(out, "%% Path: %d\n", path_num);
        fprintf(out, "%% PathStart: %d %d\n", cur->x, height - cur->y);
      }
      if (for_debug) {
        fprintf(out, "newpath\n");
        if (relative_moves) {
          fprintf(out, "0 0 moveto\n");
          lastx = lasty = 0;
        }
      }

      x = cur->x; y = height - cur->y;
      if (relative_moves) { x = x - lastx; y = y - lasty; }
      fprintf(out, "%d %d %s\n", x, y, moveto);

      lastx = cur->x; lasty = height - cur->y;
      if (for_debug) {
        fprintf(out, "gsave currentpoint 2 0 rmoveto " \
                       "2 0 360 arc fill grestore\n");
        fprintf(out, "gsave -2 3 rmoveto (%d) show grestore\n",
                     path_num);
      }
      next = cur->next;
      while(cur = next) {
        x = cur->x; y = height - cur->y;
        if (relative_lines) { x = x - lastx; y = y - lasty; }
        fprintf(out, "%d %d %s\n", x, y, lineto);
        lastx = cur->x; lasty = height - cur->y;
        next = cur->next;
      }
      if (path->closed)
        fprintf(out, "%s\n", closepath);
      if (for_debug) {
        fprintf(out, "stroke\n");
        fprintf(out, "currenthsbcolor 3 -1 roll .1 add 3 1 roll sethsbcolor\n");
      }
      if (add_comments)
        fprintf(out, "%% PathEnd: %d %d\n", lastx, lasty);
      fprintf(out, "\n");
      path_num++;
    }
    nextpl = curpl->next;
  }

  if (for_debug) fprintf(out, "stroke showpage\n");
  else           fprintf(out, "eofill showpage\n");

  fclose(out);
  return(0);
}

int read_pgm_file(char *filename, unsigned char **image,
                  int *width, int *height, int *depth)
{
  FILE *in;
  char buf[128];
  int buf_len;
  int image_size;
  int parse_state;
  if (filename) {
    if ((in = fopen(filename, "r"))==NULL) {
      snprintf(errstr, 128, "Couldn't open PGM file %s.", filename);
      return(-1);
    }
  } else in = stdin;
  for(parse_state=0; parse_state<3; parse_state++) {
    if (fgets(buf, 128, in)) {
      if (buf[0] == '#') parse_state--;
      else {
        chop_newline(buf);
        switch(parse_state) {
          case 0:
            if (strncmp(buf, "P5", 2)) {
              snprintf(errstr, 128, "Magic value from PGM input is \"%s\" " \
                                    "instead of \"P5\".", buf);
              return(-5);
            }
            break;
          case 1:
            if (sscanf(buf, "%d %d", width, height)!=2) {
              snprintf(errstr, 128, "Couldn't parse \"%s\" from PGM input " \
                                    "as width and height.", buf);
              return(-6);
            }
            break;
          case 2:
            if (sscanf(buf, "%d", depth)!=1) {
              snprintf(errstr, 128, "Couldn't parse \"%s\" from PGM input " \
                                    "as depth.", buf);
              return(-7);
            }
            break;
        }
      }
    } else {
      char *parse_target;
      switch(parse_state) {
        case 0: parse_target = "magic value"; break;
        case 1: parse_target = "width and height"; break;
        case 2: parse_target = "depth"; break;
      }
      snprintf(errstr, 128, "Failed to read the %s from the PGM input.",
                            parse_target);
      return(-1*parse_state-2);
    }
  }
  image_size = (*width) * (*height);
  if (!(*image = (uint8*) malloc (image_size))) {
    snprintf(errstr, 128, "Failed to allocate memory for image.");
    return(-8);
  }
  if ((buf_len=fread(*image, 1, image_size, in)) < image_size) {
    snprintf(errstr, 128, "Only read %d bytes (of %d bytes) from PGM input.",
                          buf_len, image_size);
    return(-9);
  }
  if (filename) fclose(in);
  return(0);
}

struct options_vals_s {
  int    min_path_len;
  int    smooth;
  double smooth_max_segment_length;
  double smooth_max_deviation;
  int    join;
  double join_radius;
  int    remove_tails;
  int    remove_tails_max_len;
  double remove_tails_neck_width;
  int    close_paths;
  double close_paths_radius;
  int    sort_paths;
  int    output_debugging;
  int    output_comments;
  int    output_minimize_size;
  int    output_relative_moves;
  int    output_relative_lines;
};
struct options_vals_s *parse_options(int argc, char *argv[]) {
  struct options_vals_s *options;
  options = (struct options_vals_s*) malloc (sizeof(struct options_vals_s));

  struct getopt_auto_option_s option_list[] =
  {
    { "min-path-len",             0, &options->min_path_len,
        1, GETOPT_AUTO_INTEGER_ARG, "Paths shorter than this will be "        \
                                    "discarded as noise"                      },
    { "smooth",                 's', &options->smooth,
        0, 0,                       "Apply the smoothing algorithm"           },
    { "smooth-max-segment-len",   0, &options->smooth_max_segment_length,
        1, GETOPT_AUTO_DOUBLE_ARG,  "Maximum length of segments that will "   \
                                    "be collapsed by the smoothing algorithm" },
    { "smooth-max-deviation",     0, &options->smooth_max_deviation,
        1, GETOPT_AUTO_DOUBLE_ARG,  "Maximum deviation of intermediate "      \
                                    "points that will be removed during "     \
                                    "the smoothing algorithm"                 },
    { "join",                   'j', &options->join,
        0, 0,                       "Apply the adjacent path joining "        \
                                    "algorithm"                               },
    { "join-radius",              0, &options->join_radius,
        1, GETOPT_AUTO_DOUBLE_ARG,  "The threshold radius to determine if "   \
                                    "two path endpoints will be joined "      \
                                    "together to form one new path including "\
                                    "all the points of the two paths"         },
    { "remove-tails",           't', &options->remove_tails,
        0, 0,                       "Apply the algorithm to remove the tails "\
                                    "from almost closed paths"                },
    { "remove-tails-max-len",     0, &options->remove_tails_max_len,
        1, GETOPT_AUTO_INTEGER_ARG, "The maximum number of segments that "    \
                                    "will be clipped as a tail" },
    { "remove-tails-neck-width",  0, &options->remove_tails_neck_width,
        1, GETOPT_AUTO_DOUBLE_ARG,  "Tails are determined by being on the "   \
                                    "open ends of paths that form an almost " \
                                    "closed loop, with a neck of no more "    \
                                    "than this width" },
    { "close-paths",            'c', &options->close_paths,
        0, 0,                       "Apply the algorithm to attempt to close "\
                                    "paths that are almost closed"            },
    { "close-paths-radius",       0, &options->close_paths_radius,
        1, GETOPT_AUTO_DOUBLE_ARG,  "The threshold radius to determine if "   \
                                    "two endpoints of a path will be joined " \
                                    "together to close the path"              },
    { "sort-paths",               0, &options->sort_paths,
        0, 0,                       "Attempt to sort the paths first by "     \
                                    "containment (paths completely "          \
                                    "surrounded by another path will be "     \
                                    "immediately after the surrounding "      \
                                    "path), then by left edge in "            \
                                    "left-to-right precedence."               },
    { "output-debugging",       'd', &options->output_debugging,
        0, 0,                       "Make the PostScript output easier to "   \
                                    "debug by marking the start of each "     \
                                    "path, and make each path a different "   \
                                    "color"                                   },
    { "output-comments",          0, &options->output_comments,
        0, 0,                       "Make the PostScript output include "     \
                                    "comments to make hand editing easier "   \
                                    "(for instance, adding path numbers to "  \
                                    "allow correlation with the debugging "   \
                                    "output, and adding as a comment the "    \
                                    "relative or absolute coordinates "       \
                                    "of moves)"                               },
    { "output-minimize",        'm', &options->output_minimize_size,
        0, 0,                       "Make the PostScript output smaller by "  \
                                    "doing things like using abbreviations "  \
                                    "for common commands"                     },
    { "output-relative-moves",    0, &options->output_relative_moves,
        0, 0,                       "Make the PostScript output use relative "\
                                    "moves, rather than absolute."            },
    { "output-relative-lines",    0, &options->output_relative_lines,
        0, 0,                       "Make the PostScript output use relative "\
                                    "lines, rather than absolute."            },
    { 0, 0, 0, 0, 0, 0 }
  };
  char *usage = "Takes a PGM file as input on the standard input " \
                "and outputs a PostScript file on the standard output.";

  /* Defaults */
  options->min_path_len = 5;
  options->smooth = 1;
  options->smooth_max_segment_length = 10000;
  options->smooth_max_deviation = 1;
  options->join = 1;
  options->join_radius = 3;
  options->remove_tails = 1;
  options->remove_tails_max_len = 3;
  options->remove_tails_neck_width = 5;
  options->close_paths = 1;
  options->close_paths_radius = 5;
  options->sort_paths = 1;
  options->output_debugging = 0;
  options->output_comments = 0;
  options->output_minimize_size = 1;
  options->output_relative_moves = 0;
  options->output_relative_lines = 1;

  if (getopt_auto(argc, argv, usage, option_list)) {
    free(options);
    options = 0;
  }

  return(options);
}

int main(int argc, char *argv[]) {
  uint8 *edge;
  int width, height, depth;
  struct path_list_s *path_list;

  /* Parse options */
  struct options_vals_s *options = parse_options(argc, argv);
  if (!options) exit(-1);

  /* Read the PGM file from STDIN */
  if (read_pgm_file(NULL, &edge, &width, &height, &depth)) {
    fprintf(stderr, "Error reading input PGM file: %s\n", errstr);
    exit(-1);
  }

  /* Follow the edges to get a list of paths */
  path_list = edge_follower(width, height, depth, edge,
                            options->min_path_len);

  /* We're done with the edge raster now, so let's get rid of it */
  free(edge);

  /* Do the path smoothing algorithm */
  if (options->smooth)
    smooth_paths(path_list, options->smooth_max_segment_length,
                            options->smooth_max_deviation);

  /* Try to join adjacent paths */
  if (options->join)
    join_paths(path_list, options->join_radius);

  /* Remove any "tails" from almost closed paths */
  if (options->remove_tails)
    remove_tails(path_list, options->remove_tails_max_len,
                            options->remove_tails_neck_width);

  /* Try to close paths that are almost closed */
  if (options->close_paths)
    close_paths(path_list, options->close_paths_radius);

  /* Sort the paths */
  if (options->sort_paths)
    sort_paths(path_list);

  /* Write the paths as PostScript on STDOUT */
  if (write_path_list_to_ps_file(NULL, width, height,
                                 options->output_debugging,
                                 options->output_comments,
                                 options->output_minimize_size,
                                 options->output_relative_moves,
                                 options->output_relative_lines,
                                 path_list))
  {
    fprintf(stderr, "Error writing output PS file: %s\n", errstr);
  }

  free_path_list(path_list);

  return(0);
}
